/* Copyright 2012-present Facebook, Inc.
 * Licensed under the Apache License, Version 2.0 */
#pragma once

#include "watchman_string.h"
#include <folly/Synchronized.h>
#include <chrono>
#include <condition_variable>
#include "thirdparty/libart/src/art.h"

namespace watchman {

/**
 * Set when this change requires a recursive scan of its children.
 *
 * If an entry is recursive, then the IO thread will stat its children too.
 *
 * PendingCollection uses this to prune unnecessary notifications: if a parent
 * entry is already flagged as requiring a recursive scan, then children can be
 * pruned.
 */
#define W_PENDING_RECURSIVE 1
/**
 * This change event came from a watcher.
 *
 * Crawler uses this to distinguish between crawler-originated events and
 * watcher-originated events.
 *
 * iothread uses this flag to detect whether cookie events were discovered via a
 * crawl or watcher.
 */
#define W_PENDING_VIA_NOTIFY 2
/**
 * Set by the IO thread when it adds new pending paths while crawling.
 *
 * Crawl-only paths do not cause PendingCollection pruning. Also affects cookie
 * discovery.
 *
 * Sort of exclusive with VIA_NOTIFY...
 */
#define W_PENDING_CRAWL_ONLY 4
/**
 * Set this flag when the watcher is desynced and may have missed filesystem
 * events. The W_PENDING_RECURSIVE flag should also be set alongside it to
 * force an recrawl of the passed in directory. Cookies will not be considered
 * when this flag is set.
 */
#define W_PENDING_IS_DESYNCED 8

/**
 * Represents a change notification from the Watcher.
 */
struct PendingChange {
  w_string path;
  std::chrono::system_clock::time_point now;
  int flags;
};

} // namespace watchman

struct watchman_pending_fs : watchman::PendingChange {
  // We own the next entry and will destroy that chain when we
  // are destroyed.
  std::shared_ptr<watchman_pending_fs> next;

  watchman_pending_fs(
      w_string path,
      std::chrono::system_clock::time_point now,
      int flags)
      : PendingChange{std::move(path), now, flags} {}

 private:
  // Only used for unlinking during pruning.
  std::weak_ptr<watchman_pending_fs> prev;
  friend class PendingChanges;
};

/**
 * Holds a linked list of watchman_pending_fs instances and a trie that
 * efficiently prunes redundant changes.
 */
class PendingChanges {
 public:
  PendingChanges() = default;
  PendingChanges(PendingChanges&&) = delete;
  PendingChanges& operator=(PendingChanges&&) = delete;

  /**
   * Erase all elements from the collection.
   */
  void clear();

  /**
   * Add a pending entry.  Will consolidate an existing entry with the same
   * name. The caller must own the collection lock.
   */
  void add(
      const w_string& path,
      std::chrono::system_clock::time_point now,
      int flags);
  void add(
      struct watchman_dir* dir,
      const char* name,
      std::chrono::system_clock::time_point now,
      int flags);

  /**
   * Merge the full contents of `chain` into this collection. They are usually
   * from a stealItems() call.
   *
   * `chain` is consumed -- the links are broken.
   */
  void append(std::shared_ptr<watchman_pending_fs> chain);

  /* Moves the head of the chain of items to the caller.
   * The tree is cleared and the caller owns the whole chain */
  std::shared_ptr<watchman_pending_fs> stealItems();

  uint32_t size() const;

 protected:
  art_tree<std::shared_ptr<watchman_pending_fs>, w_string> tree_;
  std::shared_ptr<watchman_pending_fs> pending_;

 private:
  void maybePruneObsoletedChildren(w_string path, int flags);
  inline void consolidateItem(watchman_pending_fs* p, int flags);
  bool isObsoletedByContainingDir(const w_string& path);
  inline void linkHead(std::shared_ptr<watchman_pending_fs>&& p);
  inline void unlinkItem(std::shared_ptr<watchman_pending_fs>& p);
};

class PendingCollectionBase : public PendingChanges {
 public:
  explicit PendingCollectionBase(std::condition_variable& cond);
  PendingCollectionBase(PendingCollectionBase&&) = delete;
  PendingCollectionBase& operator=(PendingCollectionBase&&) = delete;

  /**
   * Sets the pinged flag to true and wakes any waiting threads.
   */
  void ping();

  /**
   * Sets the pinged flag to false.
   * Returns true if previously pinged or PendingChanges is non-empty.
   */
  bool checkAndResetPinged();

 private:
  std::condition_variable& cond_;
  bool pinged_{false};
};

class PendingCollection
    : public folly::Synchronized<PendingCollectionBase, std::mutex> {
 public:
  PendingCollection();

  /**
   * If previously pinged or non-empty, returns a locked PendingCollectionBase.
   * Otherwise, waits up to timeoutms (or indefinitely if -1 ms) for a ping().
   *
   * pinged is set to true if pinged or non-empty. The internal pinged state is
   * always false after this call.
   */
  LockedPtr lockAndWait(std::chrono::milliseconds timeoutms, bool& pinged);

 private:
  // Notified on ping().
  std::condition_variable cond_;
};

namespace watchman {

// Since the tree has no internal knowledge about path structures, when we
// search for "foo/bar" it may return a prefix match for an existing node
// with the key "foo/bard".  We use this function to test whether the string
// exactly matches the input ("foo/bar") or whether it has a slash as the next
// character after the common prefix ("foo/bar/" as a prefix).
bool is_path_prefix(
    const char* path,
    size_t path_len,
    const char* other,
    size_t common_prefix);

inline bool is_path_prefix(const w_string& key, const w_string& root) {
  return is_path_prefix(key.data(), key.size(), root.data(), root.size());
}

} // namespace watchman
